(***********************************************************************)
(*                                                                     *)
(*                           FaCiLe                                    *)
(*                 A Functional Constraint Library                     *)
(*                                                                     *)
(*            Nicolas Barnier, Pascal Brisset, LOG, CENA               *)
(*                                                                     *)
(* Copyright 2001 CENA. All rights reserved. This file is distributed  *)
(* under the terms of the GNU Lesser General Public License.           *)
(***********************************************************************)
(* $Id: fcl_goals.mli,v 1.47 2001/06/20 16:11:29 barnier Exp $ *)

(* Module [Goals]: Building and Solving Goals *)

(*** Overview *)
(*q This module provides functions and operators to build goals that will
   control the search, i.e. mainly choose and instantiate variables.
*)

type t
  (* The type of goals. *)

(*** Creation and Solving *)
val fail : t
val success : t
  (* Failure (resp. success). Neutral element for the disjunction
     (resp. conjunction) over goals. Could be implemented as
     [create (fun () -> Stak.fail "fail")] (resp. [create (fun () -> ())]). *)
val atomic : ?name:string -> (unit -> unit) -> t
  (* [atomic (?name:"atomic") f] returns a goal calling function [f].
     [f] must take [()] as argument and return [()]. [name] default
     value is ["atomic"]. *)
val create : ?name:string -> ('a -> t) -> 'a -> t
  (* [create (?name:"create") f a] returns a goal calling [f a].
     [f] should return a goal (success to stop). [name]
     default value is ["create"]. *)
val create_rec : ?name:string -> (t -> t) -> t
  (* [create_rec (?name:"create_rec") f] returns a goal calling [f]. [f]
     takes itself as argument and should return a goal (success to stop).
     Useful to write recursive goals. [name] default value is ["create_rec"]. *)
val solve : ?control:(int -> unit) -> t -> bool
  (* [solve ?control g] solves the goal [g] and returns a success ([true]) or a
     failure ([false]). The execution can be precisely controlled
     with the [control] argument whose single argument is the number
     of bactracks since the beginning of the search. This function is called
     after every local failure:

   - it can raise [Stak.Fail] to force a failure of the search in the
     current branch (i.e. backtrack);

   - it can raise any other user exception to stop the search process;

   - it must return [unit] to continue the search; this is the default
     behavior. *)

(*** Operators *)

val (&&~) : t -> t -> t
val (||~) : t -> t -> t
  (* Conjunction and disjunction over goals. Note that these two operators
     do have the SAME PRIORITY. Goals expressions must therefore be
     carefully parenthesized to produce the expected result. *)
val forto : int -> int -> (int -> t) -> t
val fordownto : int -> int -> (int -> t) -> t
  (* [forto min max g] (resp. [fordownto min max g]) returns the
     conjunctive iteration of goal [g] on increasing (resp. decreasing)
     integers from [min] (resp. [max]) to [max] (resp. [min]). *)
val once : t -> t
  (* [once g] cuts choice points left on goal [g]. *)

val sigma : ?domain:Fcl_domain.t -> (Fcl_var.Fd.t -> t) -> t
(* [sigma ?domain fgoal] creates the goal [(fgoal v)] where [v] is a new
   variable of domain [domain]. Default domain is the largest one. It can
   be considered as an existential quantification, hence the concrete
   notation [sigma] of this function (because existential quantification can be
   seen as a generalized  disjunction). *)

(*{*)
val reset : unit -> unit
  (* _Undocumented_
     Resets the OR stack (it is not done by solve). *)
val save : t list -> Fcl_stak.level
  (* _Undocumented_
     Push a goal list on the OR stack. For advanced users. *)
(*}*)

(*** Instantiation *)
val unify : Fcl_var.Fd.t -> int -> t
(* [unify var x] instantiates variable [var] to [x]. *)

val indomain : Fcl_var.Fd.t -> t
(* Non-deterministic instanciation of a variable, by labeling its domain
   (in increasing order). *)

val instantiate : (Fcl_domain.t -> int) -> Fcl_var.Fd.t -> t
(* [instantiate choose var] Non-deterministic instanciation of a variable,
   by labeling its domain using the value returned by the [choose] function. *)

val dichotomic : Fcl_var.Fd.t -> t
(* Non-deterministic instanciation of a variable, by dichotomic recursive
   exploration of its domain. *)


(*** Goal Operations on Array of Variables *)
module Array : sig
  val foralli : ?select:('a array -> int) -> (int -> 'a -> t) -> 'a array -> t
  (* [foralli ?select g a] returns the conjunctive iteration
     of goal [g] where order is computed by the heuristic [?select] which
     must raise [Not_found] to terminate, i.e.
     [g] is successively applied to [select a] and [a.(select a)].
     Default heuristic is increasing order over indices. *)
  val forall : ?select:('a array -> int) -> ('a -> t) -> 'a array -> t
  (* [forall ?select g a] defined by [foralli ?select (fun _i x -> g x) a]. *)
  val existsi : ?select:('a array -> int) -> (int -> 'a -> t) -> 'a array -> t
  (* [existsi ?select g a] returns the disjunctive iteration
     of goal [g] where order is computed by the heuristic [?select] which
     must raise [Not_found] to terminate, i.e.
     [g] is successively applied to [select a] and [a.(select a)].
     Default heuristic is increasing order over indices. *)
  val exists : ?select:('a array -> int) -> ('a -> t) -> 'a array -> t
  (* [exists ?select g a] defined by [existsi ?select (fun _i x -> g x) a]. *)
  val choose_index : (Fcl_var.Attr.t -> Fcl_var.Attr.t -> bool) -> Fcl_var.Fd.t array -> int
  (* [choose_index order fds] returns the index of the best (minimun)
     free (not instantiated) variable in the array [fds] for the criterion
     [order]. Raises [Not_found] if all variables are bound (instantiated). *)
  val not_instantiated_fd : Fcl_var.Fd.t array -> int
  (* [not_instantiated_fd fds] returns the index of one element in [fds]
     which is not instantiated. Raises [Not_found] if all variables in array
     [fds] are bound. *)
  val labeling: Fcl_var.Fd.t array -> t
  (* Standard labeling defined by [forall indomain]. *)
end

(*** Goal Operations on List of Variables *)
module List : sig
  val forall : ?select:('a list -> 'a * 'a list) -> ('a -> t) -> 'a list -> t
  (* [forall ?select g [x1;x2;...;xn]] is [g x1 &&~ g x2 &&~ ... &&~ g xn],
     i.e. returns the conjunctive iteration of goal [g] on list [a]. *)
  val exists : ?select:('a list -> 'a * 'a list) -> ('a -> t) -> 'a list -> t
  (* [exists ?select g [x1;x2;...;xn]] is [g x1 ||~ g x2 ||~ ... ||~ g xn],
     i.e. returns the disjunctive iteration of goal [g] on list [a]. *)
  val member : Fcl_var.Fd.t -> int list -> t
(* [member v l] returns the disjunctive iteration of the instantiation of
   the variable [v] to the values in the integer list [l]. Defined by
   [fun v l -> exists (fun x -> create (fun () -> Fd.unify v x)) l]. *)
  val labeling: Fcl_var.Fd.t list -> t
  (* Standard labeling defined by [forall indomain]. *)
end

(*** Optimization *)
type bb_mode = Restart | Continue
(* Branch and bound mode. *)
val minimize : ?step:int -> ?mode:bb_mode -> t -> Fcl_var.Fd.t -> (int -> unit) -> t
 (* [minimize ?step ?mode goal cost solution] runs a Branch
    and Bound algorithm on [goal] for bound [cost], with an improvement
    of at least [step] between each solution found. With [mode] equal
    to [Restart], the search restarts from the beginning for
    every step whereas with mode [Continue] (default) the search simply
    carries on with an update of the cost constraint. [solution] is called with
    the instantiation value of [cost] (which must be instantiated by [goal])
    as argument each time a solution is found; this function can therefore
    be used to store (e.g. in a reference) the current solution. Default [step]
    is 1. [minimize] always FAILS. *)

