(***********************************************************************)
(*                                                                     *)
(*                           FaCiLe                                    *)
(*                 A Functional Constraint Library                     *)
(*                                                                     *)
(*            Nicolas Barnier, Pascal Brisset, LOG, CENA               *)
(*                                                                     *)
(* Copyright 2001 CENA. All rights reserved. This file is distributed  *)
(* under the terms of the GNU Lesser General Public License.           *)
(***********************************************************************)
(* $Id: fcl_var.ml,v 1.31 2001/05/22 14:34:17 barnier Exp $ *)

module Attr = struct
  type t = {
      dom : Fcl_domain.t Fcl_stak.ref;
      on_refine : Fcl_cstr.event; 
      on_min : Fcl_cstr.event; 
      on_max : Fcl_cstr.event; 
      on_subst : Fcl_cstr.event;
      id : int;
      name : string
    }

  let create ?(name = "") domain =
    {
      dom = Fcl_stak.ref domain;
      on_refine = Fcl_cstr.new_event (); 
      on_min = Fcl_cstr.new_event (); 
      on_max = Fcl_cstr.new_event (); 
      on_subst = Fcl_cstr.new_event ();
      id = Fcl_misc.gen_int ();
      name = name
    } 

  let dom a = Fcl_stak.get a.dom
  let fprint c a = 
    Printf.fprintf c "%s%a" (if a.name = "" then "_"^string_of_int a.id else a.name) Fcl_domain.fprint (Fcl_stak.get a.dom)
  let size a = Fcl_domain.size (Fcl_stak.get a.dom)
  let min a = Fcl_domain.min (Fcl_stak.get a.dom)
  let max a = Fcl_domain.max (Fcl_stak.get a.dom)
  let values a = Fcl_domain.values (Fcl_stak.get a.dom)
  let iter f a = Fcl_domain.iter f (Fcl_stak.get a.dom)
  let member a y = Fcl_domain.member y (Fcl_stak.get a.dom)
  let id a = a.id

(* All constraints are suspended on on_subst *)
  let constraints_number a =
    let cstrs = Fcl_cstr.registered a.on_subst in
    let cstrs = List.sort ~cmp:(fun c1 c2 -> compare c1#id c2#id) cstrs in
    let rec count n last = function
	[] -> n
      |	x::xs ->
	  let idx = x#id in
	  count (if x#is_solved || idx = last then n else n+1) idx xs in
    count 0 (-1) cstrs (* id's are positive *)

  type event = t -> Fcl_cstr.event
  let on_refine a = a.on_refine
  let on_subst a = a.on_subst
  let on_min a = a.on_min
  let on_max a = a.on_max
end


type concrete_fd = Unk of Attr.t | Val of int

module Fd = struct
  type t = concrete_fd Fcl_stak.ref
  let int i = Fcl_stak.ref (Val i)
  let create ?name domain =
    Fcl_stak.ref
      (match Fcl_domain.size domain with
	   0 -> Fcl_stak.fail "Var.Fd.create: empty initial domain"
	 | 1 -> Val (Fcl_domain.min domain)
	 | _ -> Unk (Attr.create ?name domain))

  let interval ?name min max = create ?name (Fcl_domain.interval min max)

  let array ?name n min max =
    let dom = Fcl_domain.interval min max in
    let name_elt =
      match name with
	None -> fun _i -> None
      |	Some n -> fun i -> Some (Printf.sprintf "%s_%d" n i) in
    Array.init n ~f:(fun i -> create ?name:(name_elt i) dom)

  let subst v new_v =
    match Fcl_stak.get v with
      Unk a ->
    	Fcl_debug.call 't' (fun s -> Printf.fprintf s "subst %a with %d\n" Attr.fprint a new_v);
      	if Fcl_domain.member new_v (Attr.dom a)
      	then begin
	  Fcl_stak.set v (Val new_v);
	  Fcl_cstr.schedule a.Attr.on_subst;
	  Fcl_cstr.schedule a.Attr.on_refine;
	  Fcl_cstr.schedule a.Attr.on_min;
	  Fcl_cstr.schedule a.Attr.on_max;
	  Fcl_cstr.wake_all ()
      	end
      	else Fcl_stak.fail "Var.Fd.subst"
    | Val _ ->
	Fcl_debug.fatal_error "Fd.subst: bound variable (use Fd.unify on possibly bound variable)"
    
  let value (v : t) = Fcl_stak.get v

  let refine v new_a =
    match value v with
      Unk a ->
      	let vala = Attr.dom a in
	Fcl_debug.call 't' (fun s -> Printf.fprintf s "refine _%d%a with %a\n" a.Attr.id Fcl_domain.fprint vala Fcl_domain.fprint new_a);
      	assert (Fcl_domain.included new_a vala);
      	let new_size = Fcl_domain.size new_a in
      	if new_size = 0
      	then Fcl_stak.fail "Var.Fd.refine"
      	else if new_size = Fcl_domain.size vala
      	then ()
      	else if new_size = 1
      	then subst v (Fcl_domain.min new_a)
      	else begin
      	  Fcl_stak.set a.Attr.dom new_a;
      	  let min = Fcl_domain.min vala and max = Fcl_domain.max vala in
      	  Fcl_cstr.schedule a.Attr.on_refine;
      	  if Fcl_domain.min new_a > min then Fcl_cstr.schedule a.Attr.on_min;
      	  if Fcl_domain.max new_a < max then Fcl_cstr.schedule a.Attr.on_max;
	  Fcl_cstr.wake_all ()
      	end
    | Val v ->
	if not (Fcl_domain.member v new_a) then Fcl_stak.fail "Var.Fd.refine"

  let unify (v : t) new_v =
    match Fcl_stak.get v with
      Val v -> if not (v = new_v) then Fcl_stak.fail "Var.Fd.unify"
    | Unk _a -> subst v new_v

  let unify_cstr var value =
    let update () = unify var value; true
    and delay _ = () in (* Solved when posted *)
    Fcl_cstr.create ~name:"unify_cstr" ~priority:Fcl_cstr.immediate update delay
	  
  let fprint c v =
    match Fcl_stak.get v with
      Val t -> Printf.fprintf c "%d" t
    | Unk a -> Attr.fprint c a

  let fprint_array c vs =
    let n = Array.length vs in
    Printf.fprintf c "[|";
    for i = 0 to n - 2 do
      Printf.fprintf c "%a; " fprint vs.(i)
    done;
    if n = 0 then Printf.fprintf c "|]" else
    Printf.fprintf c "%a|]" fprint vs.(n-1)

  let size v =
    match Fcl_stak.get v with
      Val _ -> 1
    | Unk a -> Attr.size a
	  
  let min v =
    match Fcl_stak.get v with
      Val t -> t
    | Unk a -> Fcl_domain.min (Fcl_stak.get a.Attr.dom)
	  
  let max v =
    match Fcl_stak.get v with
      Val t -> t
    | Unk a -> Attr.max a

  let min_max v =
    match Fcl_stak.get v with
      Val t -> (t,t)
    | Unk a -> Fcl_domain.min_max (Fcl_stak.get a.Attr.dom)
	  
  let values v =
    match Fcl_stak.get v with
      Val t -> [t]
    | Unk a -> Attr.values a
	  
  let is_var v =
    match Fcl_stak.get v with
      Val _x -> false
    | Unk _a -> true
	  
  let int_value v =
    match Fcl_stak.get v with
      Val x -> x
    | Unk a -> Fcl_debug.fatal_error ("Fd.int_value: unbound variable: " ^ a.Attr.name)
	  
  let id v =
    match Fcl_stak.get v with
      Val _x -> Fcl_debug.fatal_error "Fd.id: bound variable"
    | Unk a -> a.Attr.id

  let name v =
    match Fcl_stak.get v with
      Val _x -> Fcl_debug.fatal_error "Fd.name: bound variable"
    | Unk a -> a.Attr.name

  let iter f v =
    match Fcl_stak.get v with
      Val x -> f x
    | Unk a -> Attr.iter f a

  let member v x =
    match Fcl_stak.get v with
      Val v -> v = x
    | Unk a -> Attr.member a x

  let compare v1 v2 =
    match Fcl_stak.get v1, Fcl_stak.get v2 with
      (Val n1, Val n2) -> compare n1 n2
    | (Val _, Unk _) -> -1
    | (Unk _, Val _) -> 1
    | (Unk a1, Unk a2) -> compare (Attr.id a1) (Attr.id a2)

  let equal v1 v2 = compare v1 v2 = 0
end


let delay es x c =
  match Fd.value x with
    Val _ -> ()
  | Unk a ->
      Fcl_cstr.delay (List.map ~f:(fun e -> e a) es) c
